package earth.krakatao.protocol;

import earth.krakatao.Formatter;
import earth.krakatao.KraSocketClient;
import earth.krakatao.SocketClient;
import java.nio.charset.StandardCharsets;
import java.util.HashMap;
import java.util.UUID;
import lombok.Getter;

public class KraSocketClientProtocol {

  @Getter
  private int currentCmdIDIndex;

  @Getter
  private final HashMap<Integer, KraSocketClientProtocolMessage> sendQueueMessages;
  @Getter
  private final HashMap<Integer, KraSocketClientProtocolMessage> receivedQueueMessages;

  public KraSocketClientProtocol() {
    this.currentCmdIDIndex = 9;
    this.sendQueueMessages = new HashMap<>();
    this.receivedQueueMessages = new HashMap<>();
  }

  public int generateCmdID() {
    // TODO: check max int value 2^32
    if (this.currentCmdIDIndex < 10 || this.currentCmdIDIndex > 15000) {
      this.currentCmdIDIndex = 10;
      return this.currentCmdIDIndex;
    }

    this.currentCmdIDIndex++;

    KraSocketClient.getLogger().info("currentCmdIDIndex:" + currentCmdIDIndex);

    return this.currentCmdIDIndex;
  }

  public byte[] EncodeMessage(KraSocketClientProtocolMessage kraProtocolMessage) {
    return EncodeMessage(kraProtocolMessage.getStatus(), kraProtocolMessage.getCmdID(),
        kraProtocolMessage.getDest(), kraProtocolMessage.getUuid(),
        kraProtocolMessage.getCmdNumber(), kraProtocolMessage.getArgs());
  }

  public byte[] EncodeMessage(int status, int cmdID, int dest, UUID uuid, int cmdNumber,
      String args) {
    KraSocketClient.getLogger()
        .info("sendMessage: " + status + " " + cmdID + " " + dest + " " + cmdNumber);

    int argLen = args.length();

    byte[] raw = new byte[40 + argLen];

    raw[0] = (byte) status;

    raw[1] = (byte) cmdID;
    raw[2] = (byte) (cmdID >> 8);
    raw[3] = (byte) (cmdID >> 16);
    raw[4] = (byte) (cmdID >> 24);

    raw[5] = (byte) dest;

    byte[] uuidBytes = Formatter.formatUuid(uuid).getBytes();
    //byte[] uuidBytes = Formatter.formatUuid(uuid).getBytes();

    /* for (int i = 0; i < 32; i++) {
      raw[6 + i] = uuidBytes[i];
    }*/

    System.arraycopy(uuidBytes, 0, raw, 6, 32);

    raw[38] = (byte) cmdNumber;
    raw[39] = (byte) (cmdNumber >> 8);

    if (argLen > 0) {
      for (int i = 0; i < argLen; i++) {
        raw[40 + i] = (byte) args.charAt(i);
      }
    }

    return raw;
  }

  public KraSocketClientProtocolMessage DecodeMessage(byte[] data) {
    KraSocketClient.getLogger().info("start decode");

    byte status = data[0];

    int cmdID = Byte.toUnsignedInt(data[1]);

    cmdID += Byte.toUnsignedInt(data[2]) * 256;
    cmdID += Byte.toUnsignedInt(data[3]) * 256 * 256;
    cmdID += Byte.toUnsignedInt(data[4]) * 256 * 256 * 256;

    int dest = Byte.toUnsignedInt(data[5]);

    byte[] playerUuidBytes = new byte[32];

    System.arraycopy(data, 6, playerUuidBytes, 0, 32);

    /* for (int i = 0; i < 32; i++) {
      playerUuidBytes[i] = data[6 + i];
    } */

    String playerUuid = new String(playerUuidBytes, StandardCharsets.UTF_8);

    short cmdNumber = (short) (Byte.toUnsignedInt(data[38]) + (Byte.toUnsignedInt(data[39]) * 256));

    int argLen = data.length - 40;
    byte[] argsBytes = new byte[argLen];

    /* for (int i = 0; i < argLen; i++) {
      argsBytes[i] = data[40 + i];
    } */

    System.arraycopy(data, 40, argsBytes, 0, argLen);

    String args = new String(argsBytes, StandardCharsets.UTF_8);

    KraSocketClient.getLogger()
        .info("decoded message " + status + " " + cmdID + " " + dest + " " + playerUuid + " "
            + cmdNumber + " " + args + " ");

    return new KraSocketClientProtocolMessage(status, cmdID, dest,
        Formatter.stringToUuid(playerUuid), cmdNumber, args, null);
  }

  public void removeMessageFromSendQueue(int cmdID) {
    this.sendQueueMessages.remove(cmdID);
  }

  public void removeMessageFromReceivedQueue(int cmdID) {
    this.receivedQueueMessages.remove(cmdID);
  }

  /*
   * This method is used to send a reply message to the backend that we have received the ack message.
   */
  public void replyThatWeReceivedAckMessage(SocketClient socketClient,
      KraSocketClientProtocol kraSocketClientProtocol, int cmdID, byte dest, UUID uuid,
      int cmdNumber) {
    KraSocketClientProtocolMessage kraSocketClientProtocolMessage = new KraSocketClientProtocolMessage(
        KraSocketClientProtocolStatus.REPLY.getStatus(), cmdID,
        dest, uuid,
        cmdNumber, "", null);

    socketClient.SendMessage(kraSocketClientProtocolMessage);

    kraSocketClientProtocol.removeMessageFromReceivedQueue(cmdID);
  }
}
